import os
from collections import Counter
from math import ceil, floor
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

import hdbscan
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.patches import Polygon
from sklearn.ensemble import RandomForestClassifier
from sklearn.inspection import permutation_importance
from sklearn.preprocessing import MinMaxScaler

from justclust.explore.algclusters import feature_importances


def boxplot_clusters(vname, df, cmap, qmin=0.01, qmax=0.99, fi=None):
    data = [df[c].astype(float).dropna() for c in df.columns]
    fig, ax1 = plt.subplots(figsize=(18, 6))
    # fig.canvas.manager.set_window_title('A Boxplot Example')
    fig.subplots_adjust(left=0.075, right=0.95, top=0.9, bottom=0.25)

    bp = ax1.boxplot(data, notch=False, sym="+", vert=True, whis=1.5)
    plt.setp(bp["boxes"], color="black")
    plt.setp(bp["whiskers"], color="black")
    plt.setp(bp["fliers"], color="red", marker="+")

    # Add a horizontal grid to the plot, but make it very light in color
    # so we can use it for reading data values but not be distracting
    ax1.yaxis.grid(True, linestyle="-", which="major", color="lightgrey", alpha=0.5)

    ax1.set(
        axisbelow=True,  # Hide the grid behind plot objects
        title=f"{vname} values by cluster",
        xlabel="Clusters",
        ylabel=vname,
    )

    # Now fill the boxes with desired colors
    num_boxes = len(df.columns)
    medians = np.empty(num_boxes)
    # alphas = pd.DataFrame(
    #     MinMaxScaler().fit_transform(fi), index=fi.index, columns=fi.columns
    # )
    for i in range(num_boxes):
        box = bp["boxes"][i]
        box_x = []
        box_y = []
        for j in range(5):
            box_x.append(box.get_xdata()[j])
            box_y.append(box.get_ydata()[j])
        box_coords = np.column_stack([box_x, box_y])

        # Alternate between Dark Khaki and Royal Blue
        ax1.add_patch(
            Polygon(
                box_coords,
                facecolor=cmap[df.columns[i]],
                # alpha=None if fi is None else alphas.at[vname, df.columns[i]],
            )
        )
        # Now draw the median lines back over what we just filled in
        med = bp["medians"][i]
        median_x = []
        median_y = []
        for j in range(2):
            median_x.append(med.get_xdata()[j])
            median_y.append(med.get_ydata()[j])
            ax1.plot(median_x, median_y, "k")
        medians[i] = median_y[0]
        # Finally, overplot the sample averages, with horizontal alignment
        # in the center of each box
        ax1.plot(
            np.average(med.get_xdata()),
            np.average(data[i]),
            color="w",
            marker="*",
            markeredgecolor="k",
        )

    # Set the axes ranges and axes labels
    ax1.set_xlim(0.5, num_boxes + 0.5)
    bot, top = ax1.get_ylim()
    qs = df.loc[:, -1:].astype(float).quantile([qmin, qmax])
    bot, top = qs.min().min(), qs.max().max()
    ax1.set_ylim(bottom=bot, top=top)
    ax1.set_xticklabels(df.columns, rotation=0, fontsize=8)

    # Due to the Y-axis scale being different across samples, it can be
    # hard to compare differences in medians across the samples. Add upper
    # X-axis tick labels with the sample medians to aid in comparison
    # (just use two decimal places of precision)
    pos = np.arange(num_boxes) + 1
    upper_medians = [f"{round(s, 2):.2f}" for s in medians]
    upper_weight = (
        [f"{round(w * 100, 2):.2f}%" for w in fi.loc[vname, :]]
        if fi is not None
        else [
            "",
        ]
        * num_boxes
    )

    for tick, label, weight in zip(
        range(num_boxes), ax1.get_xticklabels(), upper_weight
    ):
        ax1.text(
            pos[tick],
            0.95,
            upper_medians[tick],
            transform=ax1.get_xaxis_transform(),
            horizontalalignment="center",
            size="x-small",
            weight="semibold",
            color=cmap[df.columns[tick]],
        )
        ax1.text(
            pos[tick],
            0.90,
            upper_weight[tick],
            transform=ax1.get_xaxis_transform(),
            horizontalalignment="center",
            size="x-small",
            weight="semibold",
            color=cmap[df.columns[tick]],
        )

    # Finally, add a basic legend
    fig.text(
        0.90,
        0.98,
        "*",
        color="white",
        backgroundcolor="silver",
        weight="roman",
        size="medium",
    )
    fig.text(
        0.915, 0.98, " Average Value", color="black", weight="roman", size="x-small"
    )

    fig.tight_layout()
    return fig


# boxplot_clusters(vname, vdf.loc[:, -1:], cmap)

# %%
# def reshape cluster


def reshape_cluster_df(df, cluster_col, vname_col=None):
    """Reshape dataframe to make it easier
    from:

        f0  f1  ... fn cluster
    0
    1
    ...
    n

    to:

        s_id  vname cluster0 cluster1 ... clusterN
    0   0     f1
    1   1     f1
    ... ...   f1
    n   n     f1
    0   0     f2
    1   1     f2
    ... ...   f2
    n   n     f2
    ... ...   ...
    n   n   fn

    """
    if vname_col is None:
        cl_cat = sorted(df[cluster_col].drop_duplicates())
        rl = {}
        for c, cf in df.groupby(cluster_col):
            for var in cf.columns:
                if var != cluster_col:
                    if var not in rl:
                        vdf = pd.DataFrame(
                            index=df.index,
                            columns=[
                                "s_id",
                                "vname",
                            ]
                            + cl_cat,
                        )
                        vdf["s_id"] = df.index
                        vdf["vname"] = var
                        rl[var] = vdf
                    else:
                        vdf = rl[var]

                    vdf.loc[cf.index, c] = cf[var].astype(float)
        res = pd.concat(rl.values())
        return res
    else:
        rl = {}
        for (vname, cl), cf in df.groupby([vname_col, cluster_col]):
            if vname not in rl:
                vdf = dict()
                rl[vname] = vdf
            else:
                vdf = rl[vname]

            vdf[cl] = cf["val"]
        return rl
        # cl_cat = sorted(rl.keys()[0].keys())
        # return {vname: [vals[cl] for cl in cl_cat] for vname, vals in rl.items()}


# %%
# plot profiles


def plot_profiles(
    cs: pd.DataFrame,
    cl_cat: List[int],
    cmap: Dict[int, Tuple[int, int, int]],
    cdir: Path,
    x_col: str = "val",
    y_col: str = "vname",
    cl_col: str = "cl",
    ncols: int = 5,
    fi: pd.DataFrame = None,
):
    counter = 0
    fig, ax = None, None
    while counter < len(cl_cat):
        for cl, cdf in cs.groupby(cl_col):
            if fig is None:
                fig, axes = plt.subplots(
                    nrows=1, ncols=ncols, figsize=(4 * ncols, 6), sharey=True
                )
                ic = 0
                cstart = cl
            counter += 1
            ax = axes[ic]
            print(cl, cmap[cl])
            sns.boxplot(
                data=cdf,
                x=x_col,
                y=y_col,
                # RGBA
                color=tuple(
                    list(cmap[cl])
                    + [
                        1.0,
                    ]
                ),
                alpha=None if fi is None else fi.loc[:, cl],
                ax=ax,
            )
            ax.set(
                title=f"Cluster: {int(cl)}",  # f"{k} {vname} {dist}",
                xlabel="scaled values",
            )
            ax.legend(bbox_to_anchor=(1, 1), loc="upper left")
            ic += 1
            if ic == 5:
                cend = cl
                fig.tight_layout()
                fig.savefig(cdir / f"profile_{cstart:02.0f}_{cend:02.0f}.png", dpi=600)
                fig = None
                ic = 0
        if fig is not None:
            cend = cl
            fig.tight_layout()
            fig.savefig(cdir / f"profile_{cstart:02.0f}_{cend:02.0f}.png", dpi=600)


def plot_profiles2(
    cs: pd.DataFrame,
    vnames: List[str],
    cmap: Dict[int, Tuple[int, int, int]],
    cdir: Path,
    x_col: str = "val",
    y_col: str = "vname",
    cl_col: str = "cl",
    ncols: int = 5,
    fi: Optional[pd.DataFrame] = None,
):
    cl_cat = sorted(cs[cl_col].drop_duplicates())
    clf = reshape_cluster_df(cs, cluster_col=cl_col, vname_col=y_col)
    ncl = len(cl_cat)
    nrows = ceil(ncl / float(ncols))

    if fi is not None:
        alpha = pd.DataFrame(
            MinMaxScaler().fit_transform(fi), index=fi.index, columns=fi.columns
        ).round(decimals=3)

    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        figsize=(4 * ncols, 5 * nrows),
        sharex=True,
        sharey=True,
    )
    # fig.canvas.manager.set_window_title('A Boxplot Example')
    fig.subplots_adjust(left=0.075, right=0.95, top=0.9, bottom=0.25)
    for c, cl in enumerate(cl_cat):
        row = floor(c / ncols)
        col = floor(c - (row * ncols))
        print(f"cl: {cl}, row: {row}, col: {col}")
        ax1 = axes[row][col]

        data = [clf[vname][cl] for vname in vnames]

        bp = ax1.boxplot(data, notch=False, sym="+", vert=False, whis=1.5)
        plt.setp(bp["boxes"], color="black")
        plt.setp(bp["whiskers"], color="black")
        plt.setp(bp["fliers"], color="red", marker="+")

        # Add a horizontal grid to the plot, but make it very light in color
        # so we can use it for reading data values but not be distracting
        ax1.xaxis.grid(True, linestyle="-", which="major", color="lightgrey", alpha=0.5)

        ax1.set(
            axisbelow=True,  # Hide the grid behind plot objects
            title=f"cluster: {cl}",
        )

        # Now fill the boxes with desired colors
        num_boxes = len(vnames)
        medians = np.empty(num_boxes)

        # breakpoint()
        for i in range(num_boxes):
            box = bp["boxes"][i]
            box_coords = np.array(
                [(x, y) for x, y in zip(box.get_xdata(), box.get_ydata())]
            )

            # Alternate between Dark Khaki and Royal Blue
            ax1.add_patch(
                Polygon(
                    box_coords,
                    facecolor=cmap[cl],
                    alpha=None if fi is None else alpha.at[vnames[i], cl],
                )
            )
            # Now draw the median lines back over what we just filled in
            med = bp["medians"][i]
            med_coords = np.array(
                [(x, y) for x, y in zip(med.get_xdata(), med.get_ydata())]
            )
            ax1.plot(med_coords.T[0], med_coords.T[1], "k")
            medians[i] = med_coords[0, 0]

            # Finally, overplot the sample averages, with horizontal alignment
            # in the center of each box
            ax1.plot(
                np.average(data[i]),
                np.average(med_coords.T[1]),
                color="k",
                marker="*",
                markeredgecolor="k",
            )

        # Set the axes ranges and axes labels
        ax1.set_ylim(0, num_boxes + 0.5)
        ax1.set_xlim(0, 1)
        # if row == 0 and col == 0:
        # breakpoint()
        # ax1.set_yticklabels(list(vnames) * (row + 1), rotation=0, fontsize=8)
        ax1.set_yticklabels(list(vnames) * (c + 1), rotation=0, fontsize=8)

        # Due to the Y-axis scale being different across samples, it can be
        # hard to compare differences in medians across the samples. Add upper
        # X-axis tick labels with the sample medians to aid in comparison
        # (just use two decimal places of precision)
        # pos = np.arange(num_boxes) + 1
        # upper_medians = [f"{round(s, 2):.2f}" for s in medians]
        upper_weight = (
            [f"{round(w * 100, 2):5.2f}%" for w in fi.loc[:, cl]]
            if fi is not None
            else [
                "",
            ]
            * num_boxes
        )

        for tick, weight in zip(range(num_boxes), upper_weight):
            # ax1.text(
            #     pos[tick],
            #     0.95,
            #     upper_medians[tick],
            #     transform=ax1.get_xaxis_transform(),
            #     horizontalalignment="center",
            #     size="x-small",
            #     weight="semibold",
            #     color=cmap[cl],
            # )
            ax1.text(
                1.07,
                tick + 1,
                upper_weight[tick],
                transform=ax1.get_yaxis_transform(),
                horizontalalignment="center",
                verticalalignment="center",
                size=7,
                weight="semibold",
                color=cmap[cl],
            )

        # # Finally, add a basic legend
        # fig.text(
        #     0.90,
        #     0.98,
        #     "*",
        #     color="white",
        #     backgroundcolor="silver",
        #     weight="roman",
        #     size="medium",
        # )
        # fig.text(
        #     0.915, 0.98, " Average Value",
        #     color="black", weight="roman", size="x-small"
        # )

        # plt.show()
    fig.tight_layout()
    return fig


# plot_profiles(cs, cl_cat, cmap, x_col="val", y_col="vname", cl_col="cl")
# plot_profiles2(cs, fi.index, cmap, x_col="val", y_col="vname", cl_col="cl", fi=fi)


# %%
# feature importance per cl


def get_df_with_features_importances(data, labels):
    cl_cat = sorted(set(labels))
    fi = pd.concat(
        [
            feature_importances(data, labels.astype(int) == int(cl), int(cl))
            for cl in cl_cat
        ],
        axis="columns",
    )
    return fi.loc[data.columns, cl_cat]


def get_perm_importances(data, labels, n_repeats=10):
    """Dict of Dict with the following keys:

    importances_mean : ndarray of shape (n_features, )
        Mean of feature importance over `n_repeats`.
    importances_std : ndarray of shape (n_features, )
        Standard deviation over `n_repeats`.
    importances : ndarray of shape (n_features, n_repeats)
        Raw permutation importance scores.
    """
    cl_cat = sorted(set(labels))
    stats = {}
    for cl in cl_cat:
        rf = RandomForestClassifier(n_estimators=500, n_jobs=-1)
        rf.fit(data, labels)
        st = permutation_importance(
            rf,
            data,
            labels.astype(int) == int(cl),
            n_repeats=n_repeats,
            n_jobs=-1,
            # scoring=["accuracy", "precision"],
        )
        stats[cl] = st
    return stats


def extract_perm_stats(stats, vnames):
    res = {}
    for cl, st in stats.items():
        fi = pd.DataFrame(index=vnames)
        fi["mean"] = st["importances_mean"]
        fi["std"] = st["importances_std"]
        fi["min"] = np.min(st["importances"], axis=1)
        q25, q50, q75 = np.quantile(st["importances"], [0.25, 0.50, 0.75], axis=1)
        fi["25%"] = q25
        fi["50%"] = q50
        fi["75%"] = q75
        fi["max"] = np.max(st["importances"], axis=1)
        res[cl] = fi
    return res


def is_light(r, g, b, hsp_threshold=0.5):
    # HSP (Highly Sensitive Poo) equation is from http://alienryderflex.com/hsp.html
    hsp = np.sqrt(0.299 * (r * r) + 0.587 * (g * g) + 0.114 * (b * b))
    return True if hsp > hsp_threshold else False


def plot_stats(
    stdfs,
    cdir,
    cmap,
    column="mean",
    ncols=5,
    interval=True,
    fname=None,
    idx_order=None,
    xlabel="importance",
    ylabel="features",
    suptitle="Permutation features importance",
    xmin=None,
    xmax=None,
    hsp_threshold=0.5,
):
    cl_cat = stdfs.keys()
    ncl = len(cl_cat)
    nrows = ceil(ncl / float(ncols))
    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        sharex=True,
        sharey=True,
        figsize=(4 * ncols, 5 * nrows),
    )
    for c, cl in enumerate(cl_cat):
        irow = floor(c / ncols)
        icol = floor(c - (irow * ncols))
        ax = axes[irow][icol]
        light_bg = is_light(*cmap[cl], hsp_threshold=hsp_threshold)
        sdf = stdfs[cl]
        if idx_order is None:
            idx_order = sdf.index
        ax.barh(idx_order, sdf.loc[idx_order, column], label=cl, color=cmap[cl])
        if interval:
            for i, (vname, row) in enumerate(sdf.iterrows()):
                idx = list(idx_order).index(vname)
                ax.plot([row["25%"], row["75%"]], [idx, idx], "-k")
        ax.set(title=f"cluster {cl:2d}")
        if xmin is not None or xmax is not None:
            ax.set_xlim(xmin, xmax)
            for idx, val in enumerate(sdf.loc[idx_order, column].values):
                if val < xmin:
                    ax.text(
                        xmin + 0.1 * abs(xmin),
                        idx,
                        f"{val:.2f}",
                        horizontalalignment="center",
                        verticalalignment="center",
                        size=7,
                        weight="semibold",
                        color="black" if light_bg else "white",
                    )
                elif val > xmax:
                    ax.text(
                        xmax - 0.1 * xmax,
                        idx,
                        f"{val:.2f}",
                        horizontalalignment="center",
                        verticalalignment="center",
                        size=7,
                        weight="semibold",
                        color="black" if light_bg else "white",
                    )

        if icol == 0:
            ax.set(ylabel=ylabel)
        if irow == (nrows - 1):
            ax.set(xlabel=xlabel)
    fig.suptitle(suptitle)
    fig.tight_layout()
    fig.savefig(
        cdir / (f"premutation_fi_{column}.png" if fname is None else fname), dpi=600
    )


# %%


def cluster_stats(selc):
    # TODO: this function extract the stats without considering
    # unbalanced clusters, perhaps we should consider to compute
    # a cluster-weighted statistics
    df = selc.astype(float, errors="ignore")

    stats = {}
    gen = df.describe().T
    idxs = [idx for idx in gen.index if idx != "cl"]
    stats["all"] = gen.loc[idxs, :]

    rshp = reshape_cluster_df(df, cluster_col="cl")
    grp = rshp.loc[:, "vname":].astype(float, errors="ignore").groupby("vname")
    for cl in sorted(set(selc["cl"])):
        stats[cl] = grp[cl].describe()
    return stats


def normalize_clusters(stats, column="mean"):
    sts = stats.copy()
    sall = sts.pop("all")
    cl_cat = sorted(sts.keys())
    norm = {}
    for cl in cl_cat:
        n = (sts[cl][column] - sall[column]) / sall[column]
        n[~np.isfinite(n)] = np.nan
        norm[cl] = pd.DataFrame(
            n,
            index=n.index,
            columns=[
                column,
            ],
        )
    return norm


def plot_norm(norm, title, cdir=None, fname=None, cmap="vlag"):
    ncols = len(norm.columns)
    nrows = len(norm.index)
    nq = norm.quantile([0.01, 0.5, 0.99], axis="columns").T

    fig, ax = plt.subplots()
    for i, (idx, row) in enumerate(norm.iterrows()):
        ax.imshow(
            np.vstack([row.values, row.values]),
            aspect="auto",
            extent=[-0.5, ncols - 0.5, i, i + 1],
            cmap=cmap,
        )
        for j, val in enumerate(row.values):
            vmin, _, vmax = nq.loc[idx, :]
            if not np.isnan(val):
                ax.annotate(
                    f"{val:.1f}",
                    xy=(j, i + 0.5),
                    ha="center",
                    va="center",
                    color="black",
                    fontsize="x-small",
                )
    ax.set_ylim(0, nrows)

    ax.set_xticks(range(ncols))
    ax.set_xticklabels(
        norm.columns,
        ha="center",
        fontsize="small",
    )
    ax.set_ylabel(norm.columns.name)

    ax.set_yticks(0.5 + np.arange(nrows))
    ax.set_yticklabels(
        norm.index,
        fontsize="small",
    )
    ax.set_ylabel(norm.index.name)

    ax.set_title(title)

    ax.invert_yaxis()

    ax.grid(visible=False)

    fig.tight_layout()
    if fname:
        if cdir is None:
            cdir = Path(".")
        fig.savefig(cdir / fname, dpi=600)

    return fig


# %%


def reporting(
    data: pd.DataFrame,
    clname: str,
    clsts: pd.DataFrame,
    cdir: Path,
    preprocs: List[Optional[Callable]] = None,
    pal: str = "turbo",
):
    # TODO: this split work only with the HDBSCAN
    # k, perc, alg, _, mcs, ms, dist, csm = clname.split("_")
    alg, _, mcs, ms, dist, csm = clname.split("_")

    # esure that the directory exists
    os.makedirs(cdir, exist_ok=True)

    cl = clsts.loc[clsts.index, clname].astype(int)
    cl_cat = sorted(cl.drop_duplicates())

    cmap = {
        c: rgb
        for c, rgb in zip(
            cl_cat,
            [
                (0.5, 0.5, 0.5),
            ]
            + list(sns.color_palette(pal, len(cl_cat) - 1)),
        )
    }

    # smap = LinearSegmentedColormap.from_list("cturbo", [cmap[c] for c in cl_cat])
    selc = data.copy()
    selc.loc[cl.index, "cl"] = cl.astype(int)

    res = reshape_cluster_df(selc, cluster_col="cl")

    # apply all the pre-processing transformations
    preprocs = [] if preprocs is None else preprocs
    tdata = data.copy()
    # Do not run preprocessing because in the case of decomposing techniques
    # variables are not longer interpretable
    # for pre in preprocs:
    #     if pre is not None:
    #         tdata = pre.fit_transform(tdata)

    tdata = pd.DataFrame(
        tdata,
        columns=data.columns,
        index=data.index,
    )
    fi = get_df_with_features_importances(tdata, cl)

    # test permutation feature importance
    fraw = get_perm_importances(tdata, cl, n_repeats=10)
    fstats = extract_perm_stats(fraw, tdata.columns)
    plot_stats(fstats, cdir, cmap, column="mean", ncols=5, interval=True)
    plot_stats(fstats, cdir, cmap, column="50%", ncols=5, interval=True)

    stats = cluster_stats(selc)
    s_men = pd.DataFrame.from_dict(
        {k: v.loc[data.columns, "mean"].values for k, v in stats.items()},
        orient="columns",
    )
    s_men.index = data.columns
    s_men.to_excel(cdir / "comparison_mean.xlsx")
    s_med = pd.DataFrame.from_dict(
        {k: v.loc[data.columns, "50%"].values for k, v in stats.items()},
        orient="columns",
    )
    s_med.index = data.columns
    s_med.to_excel(cdir / "comparison_median.xlsx")

    n_mean = normalize_clusters(stats, column="mean")
    plot_stats(
        n_mean,
        cdir,
        cmap,
        column="mean",
        ncols=5,
        interval=False,
        fname="norm_comparison_mean.png",
        idx_order=data.columns,
        xlabel=r"$\frac{\mu^{c}_{f} - \bar{\mu}_{f}}{\bar{\mu}_{f}}$",
        ylabel="features",
        suptitle="Normalize difference between clusters",
        xmin=-3,
        xmax=3,
        hsp_threshold=0.6,
    )
    n_men = pd.DataFrame.from_dict(
        {k: v.loc[data.columns].values.T[0] for k, v in n_mean.items()},
        orient="columns",
    )
    n_men.index = data.columns
    n_men.to_excel(cdir / "comparison_normal_mean.xlsx")

    plot_norm(
        n_men,
        "Normalized mean difference",
        cdir=cdir,
        fname="comparison_normal_mean.png",
        cmap="vlag",
    )

    n_p50 = normalize_clusters(stats, column="50%")
    plot_stats(
        n_p50,
        cdir,
        cmap,
        column="50%",
        ncols=5,
        interval=False,
        fname="norm_comparison_median.png",
        idx_order=data.columns,
        xlabel=r"$\frac{x - \bar{x}}{\bar{x}}$",
        ylabel="features",
        suptitle="Normalize difference between clusters",
        xmin=-3,
        xmax=3,
        hsp_threshold=0.6,
    )
    n_med = pd.DataFrame.from_dict(
        {k: v.loc[data.columns].values.T[0] for k, v in n_p50.items()}, orient="columns"
    )
    n_med.index = data.columns
    n_med.to_excel(cdir / "comparison_normal_median.xlsx")
    plot_norm(
        n_med,
        "Normalized median difference",
        cdir=cdir,
        fname="comparison_normal_median.png",
        cmap="vlag",
    )

    fdir = cdir / "features"
    os.makedirs(fdir, exist_ok=True)
    for vname, vdf in res.groupby("vname"):
        print(vname)
        fig = boxplot_clusters(vname, vdf.loc[:, -1:], cmap, fi=fi)
        fig.tight_layout()
        fig.savefig(fdir / f'{vname.replace("/", "_")}.png', dpi=600)

    cf = selc.set_index("cl").unstack().reset_index()
    cf.columns = ["vname", "cl", "val"]

    k = len([c for c in cl_cat if c >= 0])
    print(cf["vname"].drop_duplicates())
    fig, ax = plt.subplots(figsize=(18, 6))
    sns.lineplot(data=cf, x="vname", y="val", hue="cl", palette=cmap, ax=ax)
    ax.set(title=f"{k} {dist}")
    ax.legend(bbox_to_anchor=(1, 1), loc="upper left")
    ax.set_xticks(cf.columns, rotation=45, fontsize=8)
    fig.tight_layout()
    fig.savefig(cdir / f"lines_{k}_{dist}.png", dpi=600)

    scal = pd.DataFrame(
        MinMaxScaler((0, 1)).fit_transform(data), columns=data.columns, index=data.index
    )

    scal.loc[cl.index, "cl"] = cl.astype(int)

    cs = scal.set_index("cl").unstack().reset_index()
    cs.columns = ["vname", "cl", "val"]
    # plot_profiles(cs, cl_cat, cmap, x_col="val", y_col="vname", cl_col="cl")
    fig = plot_profiles2(
        cs, fi.index, cmap, cdir=cdir, x_col="val", y_col="vname", cl_col="cl", fi=fi
    )
    fig.savefig(cdir / "profiles.png", dpi=600)

    fig, ax = plt.subplots(figsize=(12, 6))
    sns.lineplot(data=cs, x="vname", y="val", hue="cl", palette=cmap, ax=ax)
    ax.set(title=f"{k} {dist}")
    ax.legend(bbox_to_anchor=(1, 1), loc="upper left")
    ax.set_xticklabels(cs["vname"].drop_duplicates(), rotation=90, fontsize=8)
    fig.tight_layout()
    fig.savefig(cdir / f"lines_scaled_{k}_{dist}.png", dpi=600)

    fig, ax = plt.subplots()
    cntr = Counter(cl)
    counts = np.array([cntr[c] for c in cl_cat])
    clrs = [cmap[c] for c in cl_cat]
    ax.bar(cl_cat, counts, label=cl_cat, color=clrs)
    ax.set(
        title=(
            f"Cardinality {alg} mcs:{mcs.split('-')[1]}, "
            f"ms:{ms.split('-')[1]}, metric: {dist.split('-')[1]}"
        ),
        ylabel="points in cluster",
        xlabel="cluster label",
    )
    fig.savefig(cdir / "cardinality.png", dpi=600)

    clusterer = hdbscan.HDBSCAN(
        min_cluster_size=int(mcs.split("-")[1]),
        min_samples=int(ms.split("-")[1]),
        metric=dist.split("-")[1],
        cluster_selection_method=csm.split("-")[1],
    )
    clusterer.fit(tdata)

    fig, ax = plt.subplots()
    clusterer.condensed_tree_.plot(
        select_clusters=True, selection_palette=[cmap[c] for c in cl_cat], axis=ax
    )
    fig.savefig(cdir / "condensed_tree.png", dpi=600)

    # for trun in ["none", "lastp", "level"]
    fig, ax = plt.subplots()
    clusterer.single_linkage_tree_.plot(truncate_mode=None, axis=ax)
    fig.savefig(cdir / "single_linkage.png", dpi=600)

    lmap = LinearSegmentedColormap.from_list(
        "a",
        [
            (0.5, 0.5, 0.5),
        ]
        + list(sns.color_palette(pal, len(cl_cat) - 1)),
    )
    fig, ax = plt.subplots()
    clsts.plot(
        column=clname,
        cmap=lmap,
        categorical=True,
        legend=True,
        legend_kwds={
            "loc": "upper left",
            "bbox_to_anchor": (1, 1),
            "fmt": "{:3.0f}",
            "fontsize": "small",
        },
        ax=ax,
    )

    ax.set(title=clname)

    fig.savefig(cdir / f"map_{clname}.png", dpi=600)
